/* GPGPU computation of jacobi algorithm with SINGLE PRECISION!
* 
******************************************************************
* Subroutine Jacobi
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "jacobi.h"

#define THREADSPERBLOCK 256

#define U(j,i) u[(j) * cols + (i)]
#define F(j,i) f[(j) * cols + (i)]
#define UOLD(j,i) uold[(j) * cols + (i)]


/* Performs a single Jacobi iteration on the GPU
* --------------------------------------------------------
* Outer loop has to be distributed among the threads
* Inner loop shall be executed sequentially by each thread
*/

// TODO: Define this method as GPU kernel
// Kernels (called from host) cannot return a value
float oneJacobiIteration(float* u, float* uold, float* f, int rows, int cols, float ax, float ay, float b, float fRelax, float* residuals) { 

	int i,j;
	// TODO: Compute the global index within the arrays (with respect to block and thread ids)
	int idx = 

	// TODO: initialize interim array of residuals
	// check that you do not access memory that has not been allocated
	float residual = 0.0f;

	// TODO: Express the outer loop by starting the kernel with 
	// a corrsponding number of threads
	// TODO: Verify that it is not operated on the boundary elements
	// TODO: Save the interim residuals in the 'residuals' array
	for (j = 1; j <= cols - 2; j++) {
		for (i = 1; i <= cols - 2; i++)
		{
			float fLRes = ( ax * (UOLD(j, i-1) + UOLD(j, i+1))
			+ ay * (UOLD(j-1, i) + UOLD(j+1, i))
			+  b * UOLD(j, i) - F(j, i)) / b;
			U(j,i) = UOLD(j,i) - fRelax * fLRes;
			residual += fLRes * fLRes;
		}
	}
	return residual;
}


void Jacobi (int iRows, int iCols, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float *afU, float *afF, float* fResidual, int* iIterCount)
{
	float ax, ay, b, residual = 0.0f;
	int n = iCols * iRows;
	
	float *h_resArray = (float*)malloc(iRows * sizeof(float)); // array to store interim residuals
	float *d_uold, *d_u, *d_f, *d_resArray; // pointer to device

	// TODO: allocate memory on the device for u, uold and f


	// TODO: allocate memory on the device to store iterim residuals


	ax = 1.0f / (fDx * fDx);      // X-direction coef
	ay = 1.0f / (fDy * fDy);      // Y_direction coef
	b = -2.0f * (ax + ay) - fAlpha;     // Central coeff
	residual = 10.0f * fTolerance;

	// TODO: copy data from host to device


	while (*iIterCount < iIterMax && residual > fTolerance) 
	{
		residual = 0.0f; 
		
		// swap u and uold
		float *utmp;
		utmp = d_uold;
		d_uold = d_u;
		d_u = utmp;


		// TODO: Compute the number of blocks_per_grid
		// The TOTAL number of threads shall be at least 'iRows'
		dim3 threads_per_block(THREADSPERBLOCK);
		dim3 blocks_per_grid;


		// TODO: Launch the GPU kernel using the device pointers


		// TODO: Copy the iterim residuals from the device to the host to finally reduce them on the host


		// Final reduction on CPU
		residual = reduceOnCPU(h_resArray, iRows);

		(*iIterCount)++;
		residual = sqrtf(residual) / (iCols * iRows);
	} // while

	*fResidual = residual;

	// TODO: Copy final solution u from device to host



	// TODO: Free all allocated memory on the device



	free(h_resArray);
}

float reduceOnCPU(float *array, int size) {
	int i;
	float sum = 0.0f;
	for (i=0; i<size; i++) {
		sum+=array[i];
	}
	return sum;
}

// Checks whether a CUDA error occured
// If so, the error message is printed and the program exits
inline void checkErr(cudaError_t err, const char* file, const int line)
{
	if(cudaSuccess != err)
	{
		fprintf(stderr, "%s: Cuda error in line %d: %s.\n", file, line, cudaGetErrorString(err) );
		exit(-1);
	}
}
