#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include "realtime.h"
#include "jacobi.h"

#define U(j,i) data->afU[(j) * data->iCols + (i)]
#define F(j,i) data->afF[(j) * data->iCols + (i)]

char* oclLoadProgSource(const char* cFilename, const char* cPreamble, size_t* szFinalLength);

static void InitGPU(int argc, char** argv, cl_command_queue *oclCommandQueue, cl_context *oclContext, cl_kernel *oclKernel) {
	cl_int err;
	// gets the device id (if specified) to run on
	int devId = 0;
	if (argc > 1) {
		devId = atoi(argv[1]);
	}
	
	// TODO: Get an OpenCL platform
	cl_platform_id oclPlatform;


	// Specifies the device to run on
	cl_uint oclNumDevices;
	err = clGetDeviceIDs(oclPlatform, CL_DEVICE_TYPE_GPU, 0, NULL, &oclNumDevices);
	checkErr(err,__FILE__,__LINE__);
	if (devId < 0 || devId >= oclNumDevices) {
		printf("The specified device ID is not supported.\n");
		exit(-1);
	}
	printf("Running on GPU with ID %d.\n\n", devId);

	// TODO: Get a GPU device and create a context 
	cl_device_id* oclDeviceIds = (cl_device_id*) malloc (oclNumDevices * sizeof(cl_device_id));
	

	// TODO: Create the context on 'oclDevice'
	cl_device_id oclDevice = oclDeviceIds[devId];
	*oclContext = 


	// TODO: Create a command-queue on the GPU device
	*oclCommandQueue = 


	// Parses the context of the file into a const char*
	size_t src_size = 0;
	const char* source = oclLoadProgSource("./jacobi.cl", "", &src_size); // NVIDIA helper function

	// TODO: Create an OpenCL program from source code 'source' with size 'src_size'
	cl_program oclProgram = 


	// TODO: Build the program



	// TODO: Create kernel: handle to the compiled OpenCL function
	*oclKernel = 
}

/*
* setting values, init mpi, omp etc
*/
static void Init(struct JacobiData *data)
{
	/* default medium */
	data->iCols      = 2000;
	data->iRows      = 2000;
	data->fAlpha     = 0.8f;
	data->fRelax     = 1.0f;
	data->fTolerance = 1e-10f;
	data->iIterMax   = 50;
#ifdef READ_INPUT
	printf("Input n - matrix size in x direction:                 ");
	scanf("%d", &data->iCols);
	printf("\nInput m - matrix size in y direction:               ");
	scanf("%d", &data->iRows);
	printf("\nInput alpha - Helmholtz constant:                   ");
	scanf("%f", &data->fAlpha);
	printf("\nInput relax - Successive over-relaxation parameter: ");
	scanf("%f", &data->fRelax);
	printf("\nInput tol - error tolerance for iterrative solver:  ");
	scanf("%f", &data->fTolerance);
	printf("\nInput mits - Maximum iterations for solver:         ");
	scanf("%d", &data->iIterMax);
#elif defined DATA_LARGE
	data->iCols      = 7000;
	data->iRows      = 7000;
	data->fAlpha     = 0.8f;
	data->fRelax     = 1.0f;
	data->fTolerance = 1e-12f;
	data->iIterMax   = 2;

#elif defined DATA_SMALL
	data->iCols      = 200;
	data->iRows      = 200;
	data->fAlpha     = 0.8f;
	data->fRelax     = 1.0f;
	data->fTolerance = 1e-7f;
	data->iIterMax   = 1000;
#endif
	printf("\n-> matrix size: %dx%d"
	"\n-> alpha: %f"
	"\n-> relax: %f"
	"\n-> tolerance: %f"
	"\n-> #of iterations: %d \n\n",
	data->iCols, data->iRows, data->fAlpha, data->fRelax,
	data->fTolerance, data->iIterMax);


	/* memory allocation for serial & omp */
	data->afU = (float*) malloc(data->iRows * data->iCols * sizeof(float));
	data->afF = (float*) malloc(data->iRows * data->iCols * sizeof(float));

	/* calculate dx and dy */
	data->fDx = 2.0f / (data->iCols - 1);
	data->fDy = 2.0f / (data->iRows - 1);

	data->iIterCount = 0;

	return;
}

/*
* final cleanup routines
*/
void Finish(struct JacobiData * data)
{
	free (data->afU);
	free (data->afF);

	return;
}

/*
* print result summary
*/
static void PrintResults(const struct JacobiData * data)
{
	printf(" Number of iterations : %d\n", data->iIterCount);
	printf(" Residual             : %le\n", data->fResidual);
	printf(" Solution Error       : %1.12lf\n", data->fError);
	printf(" Elapsed Time         : %5.7lf\n", 
	data->fTimeStop-data->fTimeStart);
	printf(" MFlops               : %6.6lf\n", 
	0.000013 * data->iIterCount * (data->iCols - 2) * (data->iRows - 2)
	/ (data->fTimeStop - data->fTimeStart));
	return;
}

/*
* Initializes matrix
* Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)
*/
static void InitializeMatrix(struct JacobiData * data)
{
	int i, j, xx, yy, xx2, yy2;
	
	/* Initialize initial condition and RHS */
	for (j = 0; j < data->iRows; j++)
	{
		for (i = 0; i < data->iCols; i++)
		{
			xx = (int) (-1.0f + data->fDx * i);
			yy = (int) (-1.0f + data->fDy * j);

			xx2 = xx * xx;
			yy2 = yy * yy;

			U(j,i) = 0.0f;
			F(j,i) = -data->fAlpha * (1.0f - xx2) * (1.0f - yy2)
			+ 2.0f * (-2.0f + xx2 + yy2);
		}
	}
}

/*
* Checks error between numerical and exact solution
*/
static void CheckError(struct JacobiData * data)
{
	float error = 0.0f;
	int i, j;
	float xx, yy, temp;

	for (j = 0; j < data->iRows; j++)
	{
		for (i = 0; i < data->iCols; i++)
		{
			xx   = -1.0f + data->fDx * i;
			yy   = -1.0f + data->fDy * j;
			temp = U(j,i) - (1.0f - xx * xx) * (1.0f - yy * yy);
			error += temp * temp;
		}
	}

	data->fError = sqrtf(error) / (data->iCols * data->iRows);
	
	return;
}


int main (int argc, char** argv)
{
	int retVal = 0;    /* return value */

	struct JacobiData myData;
	
	cl_command_queue oclCommandQueue;
	cl_context oclContext;
	cl_kernel oclKernel;

	/* inits the GPU: sets device id (if specified) and creates context */
	InitGPU(argc, argv, &oclCommandQueue, &oclContext, &oclKernel);

	/* sets default values or reads from stdin */
	Init(&myData);

	if (myData.afU && myData.afF)
	{
		/* matrix init */
		InitializeMatrix(&myData);

		/* starting timer */
		myData.fTimeStart = GetRealTime();

		/* running calculations */
		Jacobi(myData.iRows, myData.iCols, myData.iIterMax, myData.fAlpha, myData.fRelax, myData.fTolerance, myData.fDx, myData.fDy, myData.afU, myData.afF, &(myData.fResidual), &myData.iIterCount, oclCommandQueue, oclContext, oclKernel);

		/* stopping timer */
		myData.fTimeStop = GetRealTime();

		/* error checking */
		CheckError(&myData);

		/* print result summary */
		PrintResults(&myData);
	}
	else
	{
		printf(" Memory allocation failed ...\n");
		retVal = -1;
	}

	/* cleanup */
	Finish(&myData);
	
	clReleaseKernel(oclKernel);
	clReleaseCommandQueue(oclCommandQueue);
	clReleaseContext(oclContext);
	
	return retVal;
}


/*
// *********************************************************************
// Helper functions of NVIDIA GPU Computing SDK
// *********************************************************************
* Copyright 1993-2010 NVIDIA Corporation.  All rights reserved.
*
* Please refer to the NVIDIA end user license agreement (EULA) associated
* with this source code for terms and conditions that govern your use of
* this software. Any use, reproduction, disclosure, or distribution of
* this software and related documentation outside the terms of the EULA
* is strictly prohibited.
*/

/*
* Loads a Program file and prepends the cPreamble to the code
* (compare oclUtils.h)
*
* @return the source string if succeeded, 0 otherwise
* @param cFilename        program filename
* @param cPreamble        code that is prepended to the loaded file, typically a set of #defines or a header
* @param szFinalLength    returned length of the code string
*/
char* oclLoadProgSource(const char* cFilename, const char* cPreamble, size_t* szFinalLength)
{
	// locals
	FILE* pFileStream = NULL;
	size_t szSourceLength;

	// open the OpenCL source code file
	#ifdef _WIN32   // Windows version
	if(fopen_s(&pFileStream, cFilename, "rb") != 0)
	{
		return NULL;
	}
	#else           // Linux version
	pFileStream = fopen(cFilename, "rb");
	if(pFileStream == 0)
	{
		printf("file: %s\n", cFilename);
		return NULL;
	}
	#endif

	size_t szPreambleLength = strlen(cPreamble);

	// get the length of the source code
	fseek(pFileStream, 0, SEEK_END);
	szSourceLength = ftell(pFileStream);
	fseek(pFileStream, 0, SEEK_SET);
	
	// allocate a buffer for the source code string and read it in
	char* cSourceString = (char *)malloc(szSourceLength + szPreambleLength + 1);
	memcpy(cSourceString, cPreamble, szPreambleLength);
	if (fread((cSourceString) + szPreambleLength, szSourceLength, 1, pFileStream) != 1)
	{
		fclose(pFileStream);
		free(cSourceString);
		return 0;
	}

	// close the file and return the total length of the combined (preamble + source) string
	fclose(pFileStream);
	if(szFinalLength != 0)
	{
		*szFinalLength = szSourceLength + szPreambleLength;
	}
	cSourceString[szSourceLength + szPreambleLength] = '\0';

	return cSourceString;
}
