/* GPGPU computation of jacobi algorithm with SINGLE PRECISION!
* 
******************************************************************
* Subroutine Jacobi
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/

#define U(j,i) u[(j) * cols + (i)]
#define F(j,i) f[(j) * cols + (i)]
#define UOLD(j,i) uold[(j) * cols + (i)]


/* Performs a single Jacobi iteration on the GPU
* --------------------------------------------------------
* Outer loop has to be distributed among the threads
* Inner loop shall be executed sequentially by each thread
*/

// TODO: Define this method as GPU kernel
// TODO: Add the type qualifiers to the variables
__kernel
void oneJacobiIteration(__global float* u, __global float* uold, __global float* f, int rows, int cols, float ax, float ay, float b, float fRelax, __global float* residuals) { 

	// TODO: Compute the global index within the arrays (with respect to block and thread ids)
	int idx = get_global_id(0);

	// initialize interim array of residuals
	// check that you do not access memory that has not been allocated
	if (idx < rows) residuals[idx] = 0.0f;

	// TODO: Express the outer loop by starting the kernel with 
	// a corrsponding number of threads

	// TODO: Verify that it is not operated on the boundary elements
	if (idx >= 1 && idx <= rows - 2) {
		for (int i = 1; i <= cols - 2; i++)
		{
			float fLRes = ( ax * (UOLD(idx, i-1) + UOLD(idx, i+1))
			+ ay * (UOLD(idx-1, i) + UOLD(idx+1, i))
			+  b * UOLD(idx, i) - F(idx, i)) / b;
			U(idx,i) = UOLD(idx,i) - fRelax * fLRes;
			residuals[idx] += fLRes * fLRes;
		}
	}
}
