#ifndef _JACOBI_H
#define _JACOBI_H

#include <CL/cl.h>

struct JacobiData
{
	/* input data */
	int iRows;
	int iCols;
	int iIterMax;
	float fAlpha;
	float fRelax;
	float fTolerance;
	
	/* calculated dx & dy */
	float fDx;
	float fDy;

	/* pointers to the allocated memory */
	float *afU;
	float *afF;
	
	/* start and end timestamps */
	double fTimeStart;
	double fTimeStop;

	/* calculated residual (output jacobi) */
	float fResidual;
	/* effective interation count (output jacobi) */
	int iIterCount;

	/* calculated error (output error_check) */
	float fError;
	
};

/* jacobi calculation routine */
void Jacobi (int iRows, int iCols, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float * afU, float *afF, float* fResidual, int* iIterCount, cl_command_queue oclCommandQueue, cl_context oclContext, cl_kernel oclKernel);

/* final cleanup routines */
void Finish (struct JacobiData *data);

float reduceOnCPU(float *array, int size);

// Checks OpenCL error
#ifdef WIN32
__inline void checkErr(cl_int err, const char* file, const int line);
#else
inline void checkErr(cl_int err, const char* file, const int line);
#endif

#endif

