#ifndef _JACOBI_H
#define _JACOBI_H

struct JacobiData
{
	/* input data */
	int iRows;
	int iCols;
	int iRowFirst;
	int iRowLast;
	int iIterMax;
	float fAlpha;
	float fRelax;
	float fTolerance;
	
	/* calculated dx & dy */
	float fDx;
	float fDy;

	/* pointers to the allocated memory */
	float *afU;
	float *afF;
	
	/* start and end timestamps */
	double fTimeStart;
	double fTimeStop;

	/* calculated residual (output jacobi) */
	float fResidual;
	/* effective interation count (output jacobi) */
	int iIterCount;

	/* calculated error (output error_check) */
	float fError;
	
	/* MPI-Variables */
	int iMyRank;   /* current process rank (number) */
	int iNumProcs; /* how many processes */
};

/* jacobi calculation routine */
void Jacobi (int iRows, int iCols, int iRowLast, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float *restrict afU, float *restrict afF, float* fResidual, int* iIterCount);

/* final cleanup routines */
void Finish (struct JacobiData *data);

#endif

